"""Common web components."""

from __future__ import annotations

import typing

from qtpy.QtGui import QDesktopServices
from qtpy.QtWebEngineWidgets import QWebEnginePage

if typing.TYPE_CHECKING:
    pass


class ExternalBrowserNavigationPage(QWebEnginePage):
    """Custom :class:`~QWebEnginePage` to open links in default browser."""

    def acceptNavigationRequest(self, url, _type, isMainFrame):  # pylint: disable=invalid-name
        """Open links in external browser."""
        if _type == QWebEnginePage.NavigationTypeLinkClicked:
            QDesktopServices.openUrl(url)
            return False
        return super().acceptNavigationRequest(url, _type, isMainFrame)



# -*- coding: utf-8 -*-

"""Collection of the core utilities."""

from __future__ import annotations

import functools
import re

__all__ = ['get_conda_cmd_path', 'is_conda_available', 'get_conda_info']

import html
import itertools
import json
import os
import pathlib
import sys
import typing

import yaml

from anaconda_navigator.exceptions import CustomMessageException
from anaconda_navigator.utils import extra_collections
from anaconda_navigator.utils import notifications
from . import launch
from . import solvers

if typing.TYPE_CHECKING:
    from . import types as conda_types


CONDA_PATH: str
if os.name == 'nt':
    CONDA_PATH = os.path.join('Scripts', 'conda-script.py')
else:
    CONDA_PATH = os.path.join('bin', 'conda')


def get_conda_cmd_path(*prefixes: str) -> typing.Optional[str]:
    """Check if conda is found on path."""
    commands: typing.List[str] = []

    conda_exe: str = os.environ.get('CONDA_EXE', '')
    if conda_exe:
        commands.append(conda_exe)

    prefix: str = os.path.abspath(sys.prefix)
    for prefix in itertools.chain(prefixes, [os.path.dirname(os.path.dirname(prefix)), prefix]):
        commands.append(os.path.join(prefix, CONDA_PATH))

    commands.append('conda')

    command: str
    for command in commands:
        stdout: str
        stderr: str
        error: bool
        stdout, stderr, error = launch.run_process(cmd_list=[command, '--version', '--json'])

        if error:
            continue

        if any(item.startswith('conda ') for item in (stdout, stderr)):
            return command

        if all(item == '' for item in (stdout, stderr)):  # if condarc is broken, conda may respond without any output
            return command

    return None


def is_conda_available() -> bool:
    """Check if conda is available in path."""
    return get_conda_cmd_path() is not None


class CondaRcManager:
    """Helper manager to validate and backup .condarc files."""

    __slots__ = ('__backups', '__broken', '__roots')

    ORIGINAL_NAME: typing.Final[str] = '.condarc'
    BACKUP_TEMPLATE: typing.Final[str] = '.condarc.backup_{index:d}'
    BACKUP_PATTERN: typing.Final[typing.Pattern[str]] = re.compile(r'^\.condarc\.backup_(\d+)$', re.M)

    def __init__(self) -> None:
        """Initialize new :class:`~CondaRcManager` instance."""
        self.__backups: typing.Final[typing.List[str]] = []
        self.__broken: typing.Final[typing.List[str]] = []
        self.__roots: typing.Optional[typing.Sequence[str]] = None

    @property
    def backups(self) -> typing.Sequence[str]:
        """
        Backups of original configuration files.

        Original file is usually removed if its backup is added to this list.
        """
        return self.__backups

    @property
    def broken(self) -> typing.Sequence[str]:
        """Files that have issues which Navigator is unable to fix."""
        return self.__broken

    @property
    def roots(self) -> typing.Sequence[str]:
        """Folders that may have configuration files within."""
        if self.__roots is not None:
            return self.__roots

        result: extra_collections.OrderedSet[str] = extra_collections.OrderedSet()
        result.add(os.path.expanduser('~'))

        option: typing.Optional[str] = os.environ.get('CONDA_PREFIX', None)
        if option:
            result.add(os.path.abspath(option))

        option = os.environ.get('CONDA_EXE', None)
        if option:
            result.add(str(pathlib.Path(option).parents[1].absolute()))

        try:
            stream: typing.TextIO
            with open(os.path.join(os.path.expanduser('~'), '.conda', 'environments.txt'), encoding='utf8') as stream:
                result.update(filter(bool, map(str.strip, stream)))
        except FileNotFoundError:
            pass

        self.__roots = tuple(result)
        return self.__roots

    def backup_all(self) -> None:
        """Backup all existing configuration files."""
        root: str
        for root in self.roots:
            self._create_backup(root)

    def notify(self) -> str:
        """
        Notify user about detected issues.

        This may end up with any kind of message box shown to the user.
        """
        message: str = self._prepare_message()
        if not message:
            return message
        if self.broken:
            raise CustomMessageException(message=message)
        if self.backups:
            notifications.NOTIFICATION_QUEUE.push(
                message=message,
                caption='Broken Conda configuration',
                tags=('conda', 'condarc'),
            )
        return message

    def validate(self) -> None:
        """Perform a quick validation of existing files."""
        root: str
        for root in self.roots:
            try:
                stream: typing.TextIO
                with open(os.path.join(root, self.ORIGINAL_NAME), 'r', encoding='utf8') as stream:
                    yaml.safe_load(stream)
            except FileNotFoundError:
                continue
            except yaml.YAMLError:
                self._create_backup(root)

    def _create_backup(self, root: str) -> None:
        """Backup configuration file located in `root` folder."""
        index: int = 1
        try:
            index += functools.reduce(max, map(int, self.BACKUP_PATTERN.findall('\n'.join(os.listdir(root)))), 0)
        except OSError:
            pass

        source: str = os.path.join(root, self.ORIGINAL_NAME)
        target: str = os.path.join(root, self.BACKUP_TEMPLATE.format(index=index))

        try:
            os.rename(source, target)
        except FileNotFoundError:
            pass
        except OSError:
            self.__broken.append(source)
        else:
            self.__backups.append(target)

    def _prepare_message(self) -> str:
        """
        Compile a message with details on affected configuration files.

        If nothing is affected - empty message is returned.
        """
        result: str = ''

        if self.broken:
            result += (
                f'<h1>There is a problem with your {self.ORIGINAL_NAME} files</h1>'
                f'<p>Unable to start Anaconda Navigator because of broken {self.ORIGINAL_NAME} files.</p>'
            )
        elif self.backups:
            result = f'<h1>There was a problem with your {self.ORIGINAL_NAME} files</h1>'

        value: str
        if self.backups:
            result += (
                '<hr><p>'
                f'Some {self.ORIGINAL_NAME} files were reset automatically. '
                'You can find their original content in respective backup files:'
            )
            for value in self.backups:
                result += f'<br>- {html.escape(value)}'
            result += '</p>'

        if self.broken:
            result += (
                '<hr><p>'
                f'Navigator was unable to fix next {self.ORIGINAL_NAME} files. '
                'Please check their permissions and content:'
            )
            for value in self.broken:
                result += f'<br>- {html.escape(value)}'
            result += '</p>'

        return result


def get_conda_info(allow_fixes: bool = False) -> typing.Optional['conda_types.CondaInfoOutput']:
    """Return conda info as a dictionary."""
    conda_cmd: typing.Optional[str] = get_conda_cmd_path()
    if conda_cmd is None:
        return None

    manager: CondaRcManager = CondaRcManager()
    manager.validate()
    manager.notify()

    while True:
        out: str = launch.run_process(cmd_list=[conda_cmd, 'info', '--json']).stdout

        result: typing.Union['conda_types.CondaInfoOutput', 'conda_types.CondaErrorOutput']
        try:
            result = json.loads(out)
        except (TypeError, ValueError):
            return None

        if 'error' not in result:
            return typing.cast('conda_types.CondaInfoOutput', result)

        error: 'conda_types.CondaErrorOutput' = typing.cast('conda_types.CondaErrorOutput', result)
        if allow_fixes and (solvers.POOL.solve(error=error) is not None):
            continue

        manager.backup_all()
        if not manager.notify():
            message: str = (
                '<h1>Unable to start Anaconda Navigator</h1>'
                '<p>Navigator is unable to contact conda.</p>'
                '<p>Run '
                '<span style="border: 1px solid grey; border-radius: 2px; font-family: monospace">conda info</span> '
                'to get more details on the issue.</p>'
            )
            raise CustomMessageException(message=message)



# -*- coding: utf-8 -*-

"""Helper component to show messages from the Navigator."""

from __future__ import annotations

__all__ = ['NotificationsComponent']

import typing

from anaconda_navigator.utils import notifications
from anaconda_navigator.widgets import dialogs
from . import common

if typing.TYPE_CHECKING:
    from qtpy import QtWidgets
    from anaconda_navigator.widgets import main_window


class NotificationsComponent(common.Component):
    """Component for notification management."""

    __alias__ = 'notifications'

    def __init__(self, parent: 'main_window.MainWindow') -> None:
        """Initialize new :class:`NotificationsComponent` instance."""
        super().__init__(parent=parent)
        self.__listener: typing.Final[notifications.NotificationListener] = notifications.NotificationListener(self)
        self.__listener.sig_notification.connect(self.__notification)

    def __notification(self, notification: notifications.Notification) -> None:
        """"""
        self.info(text=notification.message, title=notification.caption)

    def info(self, text: str = '', title: str = '') -> None:
        """Show a dialog with information message."""
        dialog: 'QtWidgets.QDialog' = dialogs.MessageBoxInformation(text=text, title=title, parent=self.main_window)
        dialog.setMinimumWidth(480)
        dialog.exec_()

    def setup(self, worker: typing.Any, output: typing.Any, error: str, initial: bool) -> None:
        """Perform component configuration from `conda_data`."""
        notifications.NOTIFICATION_QUEUE.attach(self.__listener)



# -*- coding: utf8 -*-

"""Utility components for telemetry."""

from __future__ import annotations

__all__ = ['ApiClient', 'Stats']

import abc
import collections.abc
import functools
import locale
import platform
import sys
import typing
import urllib.parse
from contextlib import suppress

import requests.cookies
from qtpy import QtCore
from qtpy import QtWidgets

from anaconda_navigator import __about__

if typing.TYPE_CHECKING:
    from anaconda_navigator.api import anaconda_api

T = typing.TypeVar('T')


def normalize_root_url(value: str) -> str:
    """
    Make sure that path is normalized to be used as a root URL.

    It strips down all query params, fragment and ensures that URL ends with a :code:`/`. The latter is required for the
    :func:`~urllib.parse.urljoin`.
    """
    scheme: str
    netloc: str
    path: str
    scheme, netloc, path, _, _ = urllib.parse.urlsplit(value)

    if not path.endswith('/'):
        path += '/'

    if not netloc:
        if path[:1] in '/':
            raise ValueError(f'{value!r} is not a valid root url')
        netloc, path = path.split('/', 1)

    return urllib.parse.urlunsplit((scheme or 'https', netloc, path, '', ''))


class ApiClient(metaclass=abc.ABCMeta):  # pylint: disable=too-few-public-methods
    """
    Base for API client classes.

    :param root_url: Base URL for the API.
    """

    __slots__ = ('__session', '__urls')

    def __init__(self, root_url: str) -> None:
        """Initialize new instance of a :class:`ApiClient`."""
        self.__session: typing.Final[requests.Session] = requests.Session()
        self.__urls: typing.Final[typing.Dict[str, str]] = {'': normalize_root_url(root_url)}

    @property
    def _cookies(self) -> requests.cookies.RequestsCookieJar:  # noqa: D401
        """Cookies attached to each request."""
        return self.__session.cookies

    @property
    def _headers(self) -> typing.MutableMapping[str, typing.Union[bytes, str]]:  # noqa: D401
        """Headers attached to each request."""
        return self.__session.headers

    def _request(
            self,
            method: str,
            url: str,
            *args: typing.Any,
            ignore_status: bool = False,
            **kwargs: typing.Any,
    ) -> requests.Response:
        """
        Send request to the service.

        :param method: HTTP method for the request.
        :param url: Path to the API endpoint. It should be a URL relative to the :code:`root_url`.
        :param args: Extra positional arguments to :func:`requests.request`.
        :param ignore_status: Do not raise an error if server responds with non-200 status code.
        :param kwargs: Extra keyword arguments to :func:`requests.request`.
        :return: Server response.
        """
        target: typing.Optional[str] = self.__urls.get(url, None)
        if target is None:
            target = self.__urls[url] = urllib.parse.urljoin(self.__urls[''], url)

        response: requests.Response = self.__session.request(method, target, *args, **kwargs)
        if not ignore_status:
            response.raise_for_status()
        return response


MACHINE_CLARIFICATION: typing.Final[collections.abc.Mapping[str, str]] = {
    'amd64': '64',
    'x86': '32',
    'x86_64': '64',
}
PLATFORM_CLARIFICATION: typing.Final[tuple[tuple[str, str], ...]] = (
    ('darwin', 'osx'),
    ('win32', 'win'),
)

StatsDetails = typing.TypedDict('StatsDetails', {
    'channels-count': int,
    'conda-version': str,
    'environments-count': int,
    'locale-encoding': str,
    'locale-language': str,
    'navigator-version': str,
    'os-long': str,
    'os-short': str,
    'platform': str,
    'python-long': str,
    'python-short': str,
    'qt-version': str,
    'screen-height': int,
    'screen-width': int,
})


class Stats:
    """Helper class to collect basic details about the execution environment."""

    @functools.cached_property
    def channels(self) -> set[str]:  # noqa: D401
        """Channels that Navigator works with."""
        channels: set[str] = set()
        with suppress(KeyError, TypeError, IndexError):
            for config in self._api.conda_config_and_sources().communicate()[0].values():
                channels.update(config.get('channels', {}))

        return channels

    @functools.cached_property
    def conda_version(self) -> str:  # noqa: D401
        """Conda version that Navigator works with."""
        return self._api.conda_package_version(pkg='conda', name='root') or 'unknown'

    @functools.cached_property
    def details(self) -> StatsDetails:  # noqa: D401
        """
        Collected details in a dictionary format.

        Can be used to send statistics to the server.
        """
        return {
            'channels-count': len(self.channels),
            'conda-version': self.conda_version,
            'environments-count': len(self.environments),
            'locale-encoding': self.locale_encoding,
            'locale-language': self.locale_language,
            'navigator-version': self.navigator_version,
            'os-long': self.os_long,
            'os-short': self.os_short,
            'platform': self.platform,
            'python-long': self.python_long,
            'python-short': self.python_short,
            'qt-version': self.qt_version,
            'screen-height': self.screen_height,
            'screen-width': self.screen_width,
        }

    @functools.cached_property
    def environments(self) -> list[str]:  # noqa: D401
        """Environments that Navigator works with."""
        envs: list[str] = []
        with suppress(KeyError, TypeError, IndexError):
            info = self._api.conda_info().communicate()[0]
            envs = info.get('envs', [])

        return envs

    @functools.cached_property
    def locale_encoding(self) -> str:  # noqa: D401
        """Encoding used for the current locale."""
        return self._locale[1] or 'unknown'

    @functools.cached_property
    def locale_language(self) -> str:  # noqa: D401
        """Language of the current locale."""
        return self._locale[0] or 'unknown'

    @functools.cached_property
    def navigator_version(self) -> str:  # noqa: D401
        """Version of the Navigator."""
        return __about__.__version__

    @functools.cached_property
    def os_long(self) -> str:  # noqa: D401
        """Name of the OS in a long form (including version)."""
        return f'{self.os_short} {platform.release()}'

    @functools.cached_property
    def os_short(self) -> str:  # noqa: D401
        """Name of the OS in a short form (without version)."""
        return platform.system()

    @functools.cached_property
    def platform(self) -> str:  # noqa: D401
        """Platform that Navigator runs in."""
        architecture: str = platform.machine().lower()
        architecture = MACHINE_CLARIFICATION.get(architecture, architecture)

        left: str
        right: str
        system: str = sys.platform.lower()
        for left, right in PLATFORM_CLARIFICATION:
            if system.startswith(left):
                system = right
                break

        return f'{system}-{architecture}'

    @functools.cached_property
    def pyqt_version(self) -> str:  # noqa: D401
        """Version of the PyQt used to display Navigator."""
        return (
                self._api.conda_package_version(pkg='pyqt5', name='root') or
                self._api.conda_package_version(pkg='pyqt4', name='root') or
                self._api.conda_package_version(pkg='pyqt', name='root') or
                'unknown'
        )

    @functools.cached_property
    def python_long(self) -> str:  # noqa: D401
        """Current version of the python with extra details."""
        return f'{platform.python_version()} ({platform.python_implementation()})'

    @functools.cached_property
    def python_short(self) -> str:  # noqa: D401
        """Short version of the python (without patch and extra details)."""
        return '.'.join(platform.python_version_tuple()[:2])

    @functools.cached_property
    def qt_version(self) -> str:  # noqa: D401
        """Version of the Qt that Navigator is running in."""
        try:
            return QtCore.__version__
        except AttributeError:
            return 'unknown'

    @functools.cached_property
    def screen_height(self) -> int:  # noqa: D401
        """Height of the primary screen."""
        return self._screen[1]

    @functools.cached_property
    def screen_width(self) -> int:  # noqa: D401
        """Width of the primary screen."""
        return self._screen[0]

    @functools.cached_property
    def _api(self) -> anaconda_api._AnacondaAPI:  # noqa: D401
        """Shortcut to :code:`AnacondaAPI`."""
        from anaconda_navigator.api import anaconda_api  # pylint: disable=import-outside-toplevel
        return anaconda_api.AnacondaAPI()

    @functools.cached_property
    def _locale(self) -> tuple[str | None, str | None]:  # noqa: D401
        """Collected details on current locale."""
        return locale.getlocale()

    @functools.cached_property
    def _screen(self) -> tuple[int, int]:  # noqa: D401
        """Collected details on screen dimensions."""
        try:
            widget: QtWidgets.QDesktopWidget = QtWidgets.QDesktopWidget()
            geometry: QtCore.QRect = widget.screenGeometry(widget.primaryScreen())
            return geometry.width(), geometry.height()
        except Exception:  # pylint: disable=broad-exception-caught
            return 0, 0



# -*- coding: utf-8 -*-

"""Type definitions for the :mod:`~anaconda_navigator.utils.conda`."""

from __future__ import annotations

__all__ = ()

import typing


CondaInfoOutput = typing.TypedDict('CondaInfoOutput', {
    'GID': int,
    'UID': int,
    'active_prefix': typing.Optional[str],
    'active_prefix_name': typing.Optional[str],
    'av_data_dir': str,
    'av_metadata_url_base': typing.Optional[str],
    'channels': typing.List[str],
    'conda_build_version': str,
    'conda_env_version': str,
    'conda_location': str,
    'conda_prefix': str,
    'conda_shlvl': int,
    'conda_version': str,
    'config_files': typing.List[str],
    'default_prefix': str,
    'env_vars': dict,
    'envs': typing.List[str],
    'envs_dirs': typing.List[str],
    'netrc_file': typing.Optional[str],
    'offline': bool,
    'pkgs_dirs': typing.List[str],
    'platform': str,
    'python_version': str,
    'rc_path': str,
    'requests_version': str,
    'root_prefix': str,
    'root_writable': bool,
    'site_dirs': typing.List[str],
    'sys.executable': str,
    'sys.prefix': str,
    'sys.version': str,
    'sys_rc_path': str,
    'user_agent': str,
    'user_rc_path': str,
    'virtual_pkgs': typing.List[typing.List[str]],
})


class CondaErrorOutput(typing.TypedDict):
    """Common structure of the error reported by Conda."""

    caused_by: str
    error: str
    exception_name: str
    exception_type: str
    message: str


class CondaValidationErrorOutput(CondaErrorOutput):
    """Error body of the :code:`CustomValidationError` reported by Conda."""

    parameter_name: str
    parameter_value: str
    source: str



# -*- coding: utf-8 -*-

# pylint: disable=no-name-in-module
# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Logger widget."""

import json
import os
import typing

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QApplication, QHBoxLayout, QHeaderView, QLabel, QTableWidget, QTableWidgetItem, QVBoxLayout

from anaconda_navigator.config import LOG_FILENAME, LOG_FOLDER
from anaconda_navigator.utils.logs import load_log, log_files
from anaconda_navigator.widgets import ButtonPrimary, ComboBoxBase, SpacerHorizontal, SpacerVertical
from anaconda_navigator.widgets.dialogs import DialogBase
from anaconda_navigator.widgets.helperwidgets import LineEditSearch


class LogViewerDialog(DialogBase):  # pylint: disable=too-many-instance-attributes
    """Logger widget."""
    def __init__(
        self,
        parent=None,
        log_folder=LOG_FOLDER,
        log_filename=LOG_FILENAME,
    ):
        """
        Logger widget.

        Parameters
        ----------
        log_folder: str
            Folder where logs are located
        log_filename: str
            Basic name for the rotating log files.
        """
        super().__init__(parent=parent)

        self._data = None
        self._columns: typing.Sequence[str] = [
            'time', 'level', 'module', 'method', 'line', 'path', 'message', 'exception', 'traceback',
        ]
        self._headers: typing.Sequence[str] = list(map(str.capitalize, self._columns))
        self._log_filename = log_filename
        self._log_folder = log_folder

        # Widgets
        self.label = QLabel('Select log file:')
        self.combobox = ComboBoxBase()
        self.table_logs = QTableWidget(self)
        self.button_copy = ButtonPrimary('Copy')
        self.text_search = LineEditSearch()

        # Widget setup
        self.table_logs.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)
        horizontal_header = self.table_logs.horizontalHeader()
        vertical_header = self.table_logs.verticalHeader()
        horizontal_header.setStretchLastSection(True)
        horizontal_header.setSectionResizeMode(QHeaderView.Fixed)
        vertical_header.setSectionResizeMode(QHeaderView.Fixed)

        self.table_logs.setSelectionBehavior(QTableWidget.SelectRows)
        self.table_logs.setEditTriggers(QTableWidget.NoEditTriggers)

        self.setWindowTitle('Log Viewer')
        self.setMinimumWidth(800)
        self.setMinimumHeight(500)
        self.text_search.setPlaceholderText('Search...')

        # Layouts
        top_layout = QHBoxLayout()
        top_layout.addWidget(self.label)
        top_layout.addWidget(SpacerHorizontal())
        top_layout.addWidget(self.combobox)
        top_layout.addStretch()
        top_layout.addWidget(SpacerHorizontal())
        top_layout.addWidget(self.text_search)
        top_layout.addWidget(SpacerHorizontal())
        top_layout.addWidget(self.button_copy)
        layout = QVBoxLayout()
        layout.addLayout(top_layout)
        layout.addWidget(SpacerVertical())
        layout.addWidget(self.table_logs)
        self.setLayout(layout)

        # Signals
        self.combobox.currentIndexChanged.connect(self.update_text)
        self.button_copy.clicked.connect(self.copy_item)
        self.text_search.textChanged.connect(self.filter_text)

        # Setup()
        self.setup()
        self.update_style_sheet()

    def update_style_sheet(self):
        """Update custom CSS stylesheet."""

    def setup(self):
        """Setup widget content."""
        self.combobox.clear()
        paths = log_files(
            log_folder=self._log_folder,
            log_filename=self._log_filename,
        )
        files = list(map(os.path.basename, paths))
        self.combobox.addItems(files)

    def filter_text(self):
        """Search for text in the selected log file."""
        index: int
        search: str = self.text_search.text().lower()
        for index, data in enumerate(self._data):
            if any(search in str(value).lower() for value in data.values()):
                self.table_logs.showRow(index)
            else:
                self.table_logs.hideRow(index)

    def row_data(self, row):
        """Give the current row data concatenated with spaces."""
        data = {}
        if self._data:
            length = len(self._data)
            if 0 >= row < length:
                data = self._data[row]
        return data

    def update_text(self, index):  # pylint: disable=unused-argument
        """Update logs based on combobox selection."""
        path = os.path.join(self._log_folder, self.combobox.currentText())
        self._data = load_log(path)

        self.table_logs.clear()
        self.table_logs.setSortingEnabled(False)
        self.table_logs.setRowCount(len(self._data))
        self.table_logs.setColumnCount(len(self._columns))
        self.table_logs.setHorizontalHeaderLabels(self._headers)

        for row, data in enumerate(self._data):
            for col, col_key in enumerate(self._columns):
                value = data.get(col_key, '')
                item = QTableWidgetItem(str(value))
                self.table_logs.setItem(row, col, item)

        for column_index in range(self.table_logs.columnCount()):
            self.table_logs.resizeColumnToContents(column_index)

        self.table_logs.resizeRowsToContents()
        self.table_logs.setSortingEnabled(True)
        self.table_logs.scrollToBottom()
        self.table_logs.scrollToTop()
        self.table_logs.sortByColumn(1, Qt.AscendingOrder)

        # Make sure there is always a selected row
        self.table_logs.setCurrentCell(0, 0)

    def copy_item(self):
        """Copy selected item to clipboard in markdown format."""
        app = QApplication.instance()
        items = self.table_logs.selectedIndexes()
        if items:
            rows = set(sorted(item.row() for item in items))
            if self._data:
                all_data = [self._data[row] for row in rows]
                dump = json.dumps(all_data, sort_keys=True, indent=4)
                app.clipboard().setText('```json\n' + dump + '\n```')



"""Module for SignalWatcher class."""
from __future__ import annotations

import typing
from qtpy.QtCore import QObject


class SignalWatcher(QObject):
    """A class for connecting to multiple signals."""

    def __init__(
            self,
            callback: typing.Callable[..., typing.Any],
            callback_args: typing.Tuple[typing.Any, ...] | None = None,
            callback_kwargs: typing.Mapping[str, typing.Any] | None = None
    ) -> None:
        """Initialize new :class:`~SignalWatcher`  instance."""
        super().__init__()
        self.__callback: typing.Callable[..., typing.Any] = callback
        self.__signal_control_mapping: typing.MutableMapping[str, bool] = {}

        self.__callback_args: typing.Tuple[typing.Any, ...] = ()
        self.__callback_kwargs: typing.Mapping[str, typing.Any] = {}
        self.__set_callback_arguments(callback_args, callback_kwargs)

    def __set_callback_arguments(
            self,
            callback_args: typing.Tuple[typing.Any, ...] | None,
            callback_kwargs: typing.Mapping[str, typing.Any] | None
    ) -> None:
        self.__callback_args = callback_args if callback_args else self.__callback_args
        self.__callback_kwargs = callback_kwargs if callback_kwargs else self.__callback_kwargs

    def register_signal(self, signal_alias: str) -> None:
        """Register a signal with the watcher."""
        self.__signal_control_mapping[signal_alias] = False

    def signal_received(
            self, signal_alias: str,
            signal_args: typing.Tuple[typing.Any, ...] | None = None,
            signal_kwargs: typing.Mapping[str, typing.Any] | None = None,
            propagate_callback_args: bool = False
    ) -> None:
        """Mark a signal as received and trigger the callback if all signals are received."""
        if signal_alias not in self.__signal_control_mapping:
            return

        self.__signal_control_mapping[signal_alias] = True

        if propagate_callback_args:
            self.__set_callback_arguments(signal_args, signal_kwargs)

        if all(self.__signal_control_mapping.values()):
            self.__callback(*self.__callback_args, *self.__callback_kwargs)
            self.reset()

    def reset(
            self,
            callback_args: typing.Tuple[typing.Any, ...] | None = None,
            callback_kwargs: typing.Mapping[str, typing.Any] | None = None
    ) -> None:
        """Reset received signals control mapping."""
        self.__signal_control_mapping = {}
        self.__set_callback_arguments(callback_args, callback_kwargs)
